package gg.malloc.defense;

import gg.malloc.defense.model.Game;
import gg.malloc.defense.model.Arena;
import gg.malloc.defense.model.Spawnpoint;
import gg.malloc.defense.model.Spawner;
import gg.malloc.defense.model.Wave;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.logging.Logger;

import org.bukkit.Bukkit;
import org.bukkit.World;
import org.bukkit.GameMode;
import org.bukkit.attribute.Attribute;
import org.bukkit.boss.BarColor;
import org.bukkit.boss.BarStyle;
import org.bukkit.boss.BossBar;
import org.bukkit.event.player.PlayerTeleportEvent;
import org.bukkit.event.entity.EntityDamageEvent;
import org.bukkit.entity.Entity;
import org.bukkit.entity.LivingEntity;
import org.bukkit.entity.EntityType;
import org.bukkit.entity.Player;
import org.bukkit.scheduler.BukkitTask;

public class GameRunner {
  HashSet<Entity> m_livingMobs = new HashSet<>();
  int m_createdMobs = 0;
  int m_killedMobs = 0;
  Arena m_arena;
  Game m_game;
  State m_state;
  HashSet<Player> m_players = new HashSet<>();
  HashSet<Player> m_livingPlayers = new HashSet<>();
  Plugin m_plugin;

  BossBar m_gameBar = Bukkit.createBossBar("Malloc Defense", BarColor.PURPLE, BarStyle.SOLID);
  BossBar m_waveBar = Bukkit.createBossBar("Malloc Defense", BarColor.PURPLE, BarStyle.SOLID);
  BukkitTask m_countdownTask;

  int m_currentWaveNum = 0;
  int m_currentBatch = 0;
  Wave m_currentWave = null;

  Logger m_log;

  enum State {
    Idle,
    Warmup,
    Playing,
    GameOver
  }

  public GameRunner(Plugin plugin, Game game, Arena arena) {
    m_plugin = plugin;
    m_game = game;
    m_arena = arena;
    m_state = State.Idle;
    m_gameBar.setVisible(true);
    m_waveBar.setVisible(false);
    m_log = m_plugin.getLogger();
  }

  int m_warmupCountdown = 0;

  public void handleEntityDamage(EntityDamageEvent evt) {
    Entity entity = evt.getEntity();
    m_log.info("Damage " + entity);
    m_log.info("Living Mobs " + m_livingMobs);
    if (m_livingMobs.contains(entity)) {
      m_game.onMobDamaged(entity);
    }
  }

  private void countdownTick() {
    if (m_warmupCountdown == 0) {
      requestTransition(State.Playing);
    } else {
      updateMobBars();
      broadcastMessage("Starting game in " + m_warmupCountdown);
      m_warmupCountdown--;
      m_countdownTask = m_plugin.getServer().getScheduler().runTaskLater(m_plugin, () -> {
        countdownTick();
      }, 20);
    }
  }

  private void clearMobs() {
    for(Entity e : m_livingMobs) {
      e.remove();
    }
    m_livingMobs.clear();
    m_createdMobs = 0;
    m_killedMobs = 0;
  }

  private boolean enterIdle() {
    broadcastMessage("Game state: Idle");
    m_currentWaveNum = 0;
    m_livingPlayers.clear();
    for(Player p : m_players) {
      m_livingPlayers.add(p);
    }
    if (m_countdownTask != null) {
      m_countdownTask.cancel();
      m_countdownTask = null;
    }
    clearMobs();
    return true;
  }

  private boolean enterWarmup() {
    m_log.info("Game state: Warmup");
    m_currentWaveNum += 1;
    m_currentWave = m_game.getWave(m_currentWaveNum);
    m_warmupCountdown = 10;
    for(Player p : m_players) {
      m_livingPlayers.add(p);
    }
    broadcastTitle("Warmup", "Prepare yourself for wave " + m_currentWaveNum);
    clearMobs();
    countdownTick();
    return true;
  }

  private boolean enterPlaying() {
    m_log.info("Game state: Playing");
    m_log.info("Starting wave " + m_currentWaveNum);
    m_currentBatch = 1;
    spawnNextBatch();
    broadcastTitle("Wave " + m_currentWaveNum);
    return true;
  }

  private boolean enterGameOver() {
    broadcastTitle("Game Over!");
    if (m_countdownTask != null) {
      m_countdownTask.cancel();
      m_countdownTask = null;
    }
    clearMobs();
    for(Player p : m_players) {
      removePlayer(p);
    }
    return true;
  }

  private void updateMobBars() {
    m_gameBar.setVisible(true);
    switch(m_state) {
      case Idle:
        m_gameBar.setProgress(1.0);
        m_gameBar.setTitle("Waiting for playres...");
        m_gameBar.setColor(BarColor.PURPLE);
        m_waveBar.setVisible(false);
        break;
      case Warmup:
        m_gameBar.setProgress((double)m_currentWaveNum / (double)m_game.getWaveCount());
        m_gameBar.setTitle("Wave " + m_currentWaveNum + " / " + m_game.getWaveCount());
        m_gameBar.setColor(BarColor.PURPLE);
        m_waveBar.setVisible(true);
        m_waveBar.setColor(BarColor.YELLOW);
        m_waveBar.setTitle("Warmup");
        m_waveBar.setProgress((double)m_warmupCountdown / (double)10);
        break;
      case Playing:
        m_gameBar.setProgress((double)m_currentWaveNum / (double)m_game.getWaveCount());
        m_gameBar.setTitle("Wave " + m_currentWaveNum + " / " + m_game.getWaveCount());
        m_gameBar.setColor(BarColor.PURPLE);
        if (m_createdMobs > 0) {
          m_waveBar.setVisible(true);
          m_waveBar.setColor(BarColor.GREEN);
          m_waveBar.setTitle("Mobs remaining: " + (m_createdMobs - m_killedMobs));
          m_waveBar.setProgress((double)m_killedMobs / (double)m_createdMobs);
        } else {
          m_waveBar.setVisible(false);
        }
        break;
      case GameOver:
        m_gameBar.setColor(BarColor.RED);
        m_gameBar.setProgress(1.0);
        m_gameBar.setTitle("Game Over!");
        m_waveBar.setVisible(false);
        break;
    }
  }

  private void spawnNextBatch() {
    broadcastMessage("Spawning batch " + m_currentBatch);
    Spawner spawner = new GameSpawner(m_arena.spawnpoints());
    m_currentWave.spawnBatch(spawner, m_currentBatch);
    updateMobBars();
  }

  public void handlePlayerDeath(Player player) {
    if (m_livingPlayers.contains(player)) {
      m_log.info("Player has died in game" + player);
      m_livingPlayers.remove(player);
      m_arena.getWorld().strikeLightningEffect(player.getLocation());
      player.setGameMode(GameMode.SPECTATOR);
      if (m_livingPlayers.size() == 0) {
        broadcastMessage("Everyone is dead :(");
        requestTransition(State.GameOver);
      } else {
        m_log.info("Remaining players " + m_livingPlayers.size());
      }
    }
  }

  public void handleEntityDeath(Entity entity) {
    if (m_livingMobs.contains(entity)) {
      broadcastMessage("Killed game entity " + entity);
      m_livingMobs.remove(entity);
      m_killedMobs += 1;
      updateMobBars();
      if (m_livingMobs.size() <= 3) {
        m_log.info("Starting next batch!");
        if (m_currentBatch >= m_currentWave.batchCount()) {
          if (m_currentWaveNum >= m_game.getWaveCount()) {
            requestTransition(State.GameOver);
          } else if (m_livingMobs.size() == 0) {
            requestTransition(State.Warmup);
          }
        } else {
          m_currentBatch += 1;
          spawnNextBatch();
        }
      } else {
        m_log.fine("Living mobs remaining: " + m_livingMobs.size());
      }
    }
  }

  private boolean syncPlayer(Player player) {
    m_log.fine("Synchronizing player " + player);
    World playerWorld = player.getLocation().getWorld();
    World gameWorld = m_arena.getWorld();
    m_gameBar.addPlayer(player);
    m_waveBar.addPlayer(player);
    if (m_livingPlayers.contains(player)) {
      m_log.fine("Player is alive, turning into adventure mode " + player);
      if (player.getGameMode() != Bukkit.getDefaultGameMode()) {
        player.setGameMode(Bukkit.getDefaultGameMode());
        player.teleport(gameWorld.getSpawnLocation(), PlayerTeleportEvent.TeleportCause.PLUGIN);
      }
      player.setHealth(player.getAttribute(Attribute.GENERIC_MAX_HEALTH).getValue());
      player.setFoodLevel(10);
    } else {
      m_log.fine("Player is dead, turning into spectator " + player);
      player.setGameMode(GameMode.SPECTATOR);
    }
    if (playerWorld != gameWorld) {
      m_log.info("Teleporting player " + player);
      return player.teleport(gameWorld.getSpawnLocation(), PlayerTeleportEvent.TeleportCause.PLUGIN);
    } else {
      return false;
    }
  }

  private boolean attemptTransition(State state) {
    if (m_state == state) {
      return false;
    }
    switch(state) {
      case Idle:
        return enterIdle();
      case Warmup:
        return enterWarmup();
      case Playing:
        return enterPlaying();
      case GameOver:
        return enterGameOver();
    }
    return false;
  }
  
  private boolean validateTransition(State from, State to) {
    switch(from) {
      case Idle:
        return to == State.Warmup;
      case Warmup:
        return to == State.Playing || to == State.Idle || to == State.GameOver;
      case Playing:
        return to == State.Warmup || to == State.GameOver || to == State.Idle;
      case GameOver:
        return to == State.Idle;
    }
    return false;
  }

  public boolean requestTransition(State state) {
    if (!validateTransition(m_state, state)) {
      m_log.warning("Attemped illegal transition: " + m_state + " -> " + state);
      return false;
    }
    if (attemptTransition(state)) {
      updateMobBars();
      m_log.info("Game transition: " + m_state + " -> " + state);
      m_state = state;
      for(Player p : m_players) {
        syncPlayer(p);
      }
      return true;
    }
    m_log.severe("Failed to complete transition: " + m_state + " -> " + state);
    return false;
  }

  public boolean addPlayer(Player p) {
    if (m_state == State.Idle || m_state == State.Warmup) {
      m_log.info("Adding player " + p);
      m_players.add(p);
      broadcastMessage(p.getName() + " has joined the game");
      syncPlayer(p);
      if (m_state == State.Idle) {
        requestTransition(State.Warmup);
      }
      return true;
    } else {
      return false;
    }
  }

  public void removePlayer(Player p) {
    m_log.info("Removing player " + p); 
    m_gameBar.removePlayer(p);
    m_waveBar.removePlayer(p);
    m_players.remove(p);
    m_livingPlayers.remove(p);
    p.setGameMode(Bukkit.getDefaultGameMode());
    if (m_players.size() == 0) {
      requestTransition(State.Idle);
    }
  }

  void broadcastTitle(String title) {
    broadcastTitle(title, "");
  }

  void broadcastTitle(String title, String subtitle) {
    for(Player p : m_players) {
      p.sendTitle(title, subtitle, 10, 70, 20);
    }
  }

  void broadcastMessage(String string) {
    World world = m_arena.getWorld();
    for(Player p : world.getPlayers()) {
      p.sendMessage(string);
    }
  }

  void registerSpawnedMob(Entity entity) {
    m_log.fine("Registered new mob " + entity); 
    m_livingMobs.add(entity);
    m_createdMobs += 1;
  }

  private class GameSpawner implements Spawner {
    Spawnpoint[] m_spawnpoints;
    int m_spawnIdx = 0;
 
    public GameSpawner(Spawnpoint[] spawnpoints) {
      m_spawnpoints = spawnpoints;
    }

    @Override
    public Entity spawnMob(EntityType type) {
      m_log.fine("Spawning " + type + " at " + m_spawnpoints[m_spawnIdx]);
      Entity newMob = m_arena.getWorld().spawnEntity(m_spawnpoints[m_spawnIdx].getLocation(), type);
      LivingEntity livingMob = (LivingEntity)newMob;
      livingMob.setRemoveWhenFarAway(false);
      registerSpawnedMob(newMob);
      m_spawnIdx += 1;
      m_spawnIdx %= m_spawnpoints.length;
      return newMob;
    }
  }
}
